import pandas as pd; import numpy as np
from pyg_base import drange, eq
from pyg_timeseries import cumprod, cumprod_, cumsum, cumsum_, diff, diff_, ewma, ewma_, ewmcor, ewmcor_,\
    ewmvar, ewmvar_, ewmrms, ewmrms_, ewmskew, ewmskew_, ewmstd, ewmstd_, \
    expanding_max, expanding_max_, expanding_mean, expanding_mean_, expanding_min, expanding_min_, expanding_rms, expanding_rms_, expanding_skew, expanding_skew_, expanding_std, expanding_std_, expanding_sum, expanding_sum_, \
    ratio, ratio_, rolling_max, rolling_max_, rolling_mean, rolling_mean_, rolling_median, rolling_median_, rolling_min, rolling_min_, \
    rolling_rank, rolling_rank_, rolling_rms, rolling_rms_, rolling_skew, rolling_skew_, rolling_std, rolling_std_, rolling_sum, rolling_sum_, \
    shift, shift_, rolling_quantile, rolling_quantile_, \
    ts_count, ts_count_, ts_max, ts_max_, ts_mean, ts_mean_, ts_min, ts_min_, ts_rms, ts_rms_, ts_skew, ts_skew_, ts_std, ts_std_, ts_sum, ts_sum_

_data = 'data'

def _nona_data():
    s = pd.Series(np.random.normal(0,1,1000), drange(-999))
    df = pd.DataFrame(np.random.normal(0,1,(1000, 20)), drange(-999))
    return [s,df]


s, df = _nona_data()
s1 = s.iloc[:500]; s2 = s.iloc[500:]
df1 = df.iloc[:500]; df2 = df.iloc[500:]
t, ef = _nona_data()
t1 = t.iloc[:500]; t2 = t.iloc[500:]
ef1 = ef.iloc[:500]; ef2 = ef.iloc[500:]

#### here we test two things (1) that f_().data == f() and that (2) We can run additional data using the state we got from f_ run over history.

n2f = dict(cumprod = cumprod, cumsum = cumsum, diff = diff, ewma = ewma, ewmrms = ewmrms, ewmskew = ewmskew, ewmstd = ewmstd, 
           expanding_max = expanding_max, expanding_mean = expanding_mean, expanding_min = expanding_min, expanding_rms = expanding_rms, expanding_skew = expanding_skew, expanding_std = expanding_std, expanding_sum = expanding_sum, ratio = ratio, 
           rolling_max = rolling_max, rolling_mean = rolling_mean, rolling_median = rolling_median, rolling_min = rolling_min, rolling_rank = rolling_rank, rolling_rms = rolling_rms, rolling_skew = rolling_skew, rolling_std = rolling_std, rolling_sum = rolling_sum, 
           shift = shift, ts_count = ts_count, ts_max = ts_max, ts_mean = ts_mean, ts_min = ts_min, ts_rms = ts_rms, ts_skew = ts_skew, ts_std = ts_std, ts_sum = ts_sum)

n2f_ = dict(cumprod = cumprod_, cumsum = cumsum_, diff = diff_, ewma = ewma_, ewmrms = ewmrms_, ewmskew = ewmskew_, ewmstd = ewmstd_, 
            expanding_max = expanding_max_, expanding_mean = expanding_mean_, expanding_min = expanding_min_, expanding_rms = expanding_rms_, expanding_skew = expanding_skew_, expanding_std = expanding_std_, expanding_sum = expanding_sum_, ratio = ratio_, 
            rolling_max = rolling_max_, rolling_mean = rolling_mean_, rolling_median = rolling_median_, rolling_min = rolling_min_, rolling_rank = rolling_rank_, rolling_rms = rolling_rms_, rolling_skew = rolling_skew_, rolling_std = rolling_std_, rolling_sum = rolling_sum_, 
            shift = shift_, ts_count = ts_count_, ts_max = ts_max_, ts_mean = ts_mean_, ts_min = ts_min_, ts_rms = ts_rms_, ts_skew = ts_skew_, ts_std = ts_std_, ts_sum = ts_sum_)


def test_ts_states():
    n2f = dict(ts_count = ts_count, ts_max = ts_max, ts_mean = ts_mean, ts_min = ts_min, ts_rms = ts_rms, ts_skew = ts_skew, ts_std = ts_std, ts_sum = ts_sum)
    n2f_ = dict(ts_count = ts_count_, ts_max = ts_max_, ts_mean = ts_mean_, ts_min = ts_min_, ts_rms = ts_rms_, ts_skew = ts_skew_, ts_std = ts_std_, ts_sum = ts_sum_)

    for n in n2f:
        f = n2f[n]
        f_ = n2f_[n]
        res = f(s)
        res_ = f_(s)
        assert eq(res, res_.data)
        res1 = f_(s1)
        res2 = f_(s2, instate = res1.state)                    
        assert eq(res2 - _data, res_ - _data)
        assert eq(res2.data, res)

    for n in n2f:
        f = n2f[n]
        f_ = n2f_[n]
        res = f(df)
        res_ = f_(df)
        assert eq(res, res_.data)
        res1 = f_(df1)
        res2 = f_(df2, instate = res1.state)                    
        assert eq(res2 - _data, res_ - _data)
        assert eq(res2.data, res)


def test_window_double_states():
    n2f = dict(ewmcor = ewmcor) 
    n2f_ = dict(ewmcor = ewmcor_)
    for i in (5,10):
        for n in n2f:
            f = n2f[n]
            f_ = n2f_[n]
            res = f(s, t, i)
            res_ = f_(s , t, i)
            assert eq(res, res_.data)
            res1 = f_(s1, t1, i)
            res2 = f_(s2, t2, i, instate = res1.state)                    
            assert eq(res2 - _data, res_ - _data)
            assert eq(res2.data, res.iloc[500:])

    for i in (5, 10):
        for n in n2f:
            f = n2f[n]
            f_ = n2f_[n]
            res = f(df, ef, i)
            res_ = f_(a = df , b = ef, n = i)
            assert eq(res, res_.data)
            res1 = f_(df1, ef1, i)
            res2 = f_(df2, ef2, i, instate = res1.state)                    
            assert eq(res2 - _data, res_ - _data)
            assert eq(res2.data, res.iloc[500:])


def test_window_states():
    n2f = dict(ewma = ewma, ewmrms = ewmrms, ewmskew = ewmskew, ewmvar = ewmvar, ewmstd = ewmstd, diff = diff, shift = shift, ratio = ratio,
               rolling_max = rolling_max, rolling_mean = rolling_mean, rolling_median = rolling_median, rolling_min = rolling_min, 
               rolling_quantile = rolling_quantile, rolling_rank = rolling_rank, rolling_rms = rolling_rms, rolling_skew = rolling_skew, rolling_std = rolling_std, rolling_sum = rolling_sum)
    
    n2f_ = dict(ewma = ewma_, ewmrms = ewmrms_, ewmvar = ewmvar_, ewmskew = ewmskew_, ewmstd = ewmstd_, diff = diff_, shift = shift_, ratio = ratio_,
                rolling_max = rolling_max_, rolling_mean = rolling_mean_, rolling_median = rolling_median_, rolling_min = rolling_min_, 
                rolling_quantile = rolling_quantile_, rolling_rank = rolling_rank_, rolling_rms = rolling_rms_, rolling_skew = rolling_skew_, rolling_std = rolling_std_, rolling_sum = rolling_sum_)

    for i in (5, 10):
        for n in n2f:
            f = n2f[n]
            f_ = n2f_[n]
            res = f(s, i)
            res_ = f_(s , i)
            assert eq(res, res_.data)
            res1 = f_(s1, i)
            res2 = f_(s2, i, instate = res1.state)                    
            assert eq(res2 - _data, res_ - _data)
            assert eq(res2.data, res.iloc[500:])

    for i in (5, 10):
        for n in n2f:
            f = n2f[n]
            f_ = n2f_[n]
            res = f(df, i)
            res_ = f_(df , i)
            assert eq(res, res_.data)
            res1 = f_(df1, i)
            res2 = f_(df2, i, instate = res1.state)                    
            assert eq(res2 - _data, res_ - _data)
            assert eq(res2.data, res.iloc[500:])



def test_expanding_states():
    n2f = dict(expanding_max = expanding_max, expanding_mean = expanding_mean, expanding_min = expanding_min, expanding_rms = expanding_rms, expanding_skew = expanding_skew, expanding_std = expanding_std, expanding_sum = expanding_sum, cumprod = cumprod, cumsum = cumsum) 
    n2f_ = dict(expanding_max = expanding_max_, expanding_mean = expanding_mean_, expanding_min = expanding_min_, expanding_rms = expanding_rms_, expanding_skew = expanding_skew_, expanding_std = expanding_std_, expanding_sum = expanding_sum_, cumprod = cumprod_, cumsum = cumsum_)

    for n in n2f:
        f = n2f[n]
        f_ = n2f_[n]
        res = f(s)
        res_ = f_(s)
        assert eq(res, res_.data)
        res1 = f_(s1)
        res2 = f_(s2, instate = res1.state)                    
        assert eq(res2 - _data, res_ - _data)
        assert eq(res2.data, res.iloc[500:])

    for n in n2f:
        f = n2f[n]
        f_ = n2f_[n]
        res = f(df)
        res_ = f_(df)
        assert eq(res, res_.data)
        res1 = f_(df1)
        res2 = f_(df2, instate = res1.state)                    
        assert eq(res2 - _data, res_ - _data)
        assert eq(res2.data, res.iloc[500:])


def test_min_max_no_data():
    a1 = np.array([np.nan, np.nan])
    a2 = np.array([np.nan, 1., 2.])
    m1 = ts_max_(a1)
    assert np.isnan(m1.data)
    m2 = ts_max(a2, **m1)    
    assert m2 == 2
    m1 = ts_min_(a1)        
    assert np.isnan(m1.data)
    m2 = ts_min(a2, **m1)    
    assert m2 == 1



from pyg_timeseries import rolling_quantile, rolling_quantile_, fnna, expanding_rank
from pyg_base import eq, dictable, drange, nona, shape
import pandas as pd; import numpy as np
import pytest

def test_rolling_quantile_short_seq():
    a = np.random.normal(0, 1, 50)
    assert eq(rolling_quantile(a, 100), a + np.nan)


def test_rolling_quantile_reverse():
    a = np.random.normal(0, 1, 20)
    fwd = rolling_quantile(a, 10).T[0]
    bck = rolling_quantile(a, -10).T[0]
    assert eq(nona(fwd), nona(bck))
    assert fnna(fwd) == 9 and fnna(fwd, -1) == 19
    assert fnna(bck) == 0 and fnna(bck, -1) == 10


def test_rolling_quantile_fail_multicol():
    a = np.random.normal(0, 1, (20,3))
    with pytest.raises(ValueError):
        rolling_quantile(a, 10, [0.1,0.2,0.3])
        
    rolling_quantile(a, 10, [0.1])

def test_shape():
    assert shape(5) == ()
    assert shape([5]) == ()
    assert shape(dictable(a = [1,2,3])) == (3,1)
    assert shape(pd.DataFrame(dictable(a = [1,2,3]))) == (3,1)
    assert shape(pd.Series(dictable(a = [1,2,3]))) == (1,)


def test_expandning_rank():
    a = pd.Series([1.,2., np.nan, 0.,4.,2.], drange(-5)) 
    rank = expanding_rank(a) 
    assert eq(rank, pd.Series([0, 1, np.nan, -1, 1, 0.25], drange(-5))) 
    


from pyg_timeseries import ewma, ewmstd, ewmrms, ewmskew, ewmLR, ewmcor, ewmcorr, ts_max
from pyg_base import eq, dt, calendar, drange, Dict
import pandas as pd; import numpy as np

t = dt(2021,3,1)
cal = calendar('US')

def test_ewm_monthly():    
    months = drange('-500m', t, '1m')
    a = pd.Series(np.random.normal(0,1,501), months)
    days = cal.drange('-500m', t, 1)
    for f in ewma, ewmstd, ewmskew, ewmrms:
        assert eq(f(a, 3).reindex(days), f(a.reindex(days), 3))
        assert eq(f(a.reindex(days).ffill(), 3, time = 'm'), f(a.reindex(days),3).ffill())


def test_ewm_weekly():    
    months = drange('-500w', t, '1w')
    a = pd.Series(np.random.normal(0,1,501), months)
    days = cal.drange('-500w', t, '1b')
    for f in ewma, ewmstd, ewmskew, ewmrms:
        assert eq(f(a, 3).reindex(days), f(a.reindex(days), 3))
        assert eq(f(a.reindex(days).ffill(), 3, time = 'w'), f(a.reindex(days),3).ffill())

def test_ewm_yearly():    
    t = dt(2021,1,1)
    yearly = drange('-100y', t, '1y')
    a = pd.Series(np.random.normal(0,1,101), yearly)
    days = cal.drange('-100y', t, 1)
    for f in ewma, ewmstd, ewmskew, ewmrms:
        assert eq(f(a, 3).reindex(days), f(a.reindex(days), 3))
        assert eq(f(a.reindex(days).ffill(), 3, time = 'y'), f(a.reindex(days),3).ffill())

def test_ewm_LR():
    a0 = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    a1 = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    b = (a0 - a1) + pd.Series(np.random.normal(0,1,10000), drange(-9999))
    a = pd.concat([a0,a1], axis=1)
    LR = ewmLR(a,b,50)
    assert abs(LR.m.mean()[0]-1)<0.5
    assert abs(LR.m.mean()[1]+1)<0.5
    a = Dict(a0 = a0, a1 = a1)
    LR2 = ewmLR(a,b,50)
    assert eq(LR2.a0.m, LR.m[0])
    assert eq(LR2.a0.c, LR.c[0])
    assert 'state' not in LR2.a0
    a = [a0,a1]
    LR3 = ewmLR(a,b,50)
    assert eq(LR2.a0, LR3[0])
    assert eq(LR2.a1, LR3[1])

def test_ewmcor():
    """ the two functions are not identical but should be pretty close """
    np.random.seed(0)
    a = np.random.normal(0,1,(1000,10)) 
    a = a + np.random.normal(0, 1, (1000, 1))
    res = ewmcorr(a, 20)
    for i in range(10):
        for j in range(i):        
            x = ewmcor(a[:,i], a[:, j], 20)
            y = res[:, i, j]
            assert ts_max(abs(x - y)[20:])<0.05

    
def test_ewm_empty():
    for a in (np.array([]), pd.Series([],[], dtype = float), pd.DataFrame([],[])):
        for f in [ewma, ewmstd, ewmrms, ewmskew]:
            res = f(a, 20)
            assert eq(res, a)



import numpy as np
import pandas as pd
from pyg_base import is_pd, is_num, is_nums, is_list, as_list, pd2np, reducer
from pyg_timeseries._decorators import mask_nans, apply_along_first_axis
from pyg_timeseries._rolling import v2na


def _mult(a, b):
    """
    A much more forgiving np.matmul implementation.
    - supports multiplying by numbers
    - supports multiplying by vectors posing as 1xn matrices
    """
    if is_pd(a):
        a = a.values
    if is_pd(b):
        b = b.values
    if is_num(a) or is_num(b):
        return a * b
    if len(a.shape) == 1 and len(b.shape) == 1:
        return np.matmul(a,b)
    if len(a.shape) == 1:
        a = np.array([a])
    if len(b.shape) == 1:
        b = np.array([b])
    if len(a.shape) == 2 and len(b.shape) == 2 and a.shape[1]!=b.shape[0]: 
        ## we first flip and fake vectors
        if a.shape[1] == 1:
            a = a.T
        if b.shape[0] == 1:
            b = b.T
    if len(a.shape) == 2 and len(b.shape) == 2 and a.shape[1]!=b.shape[0] and a.shape[0] == b.shape[0]:
        res = np.array([_mult(a[i], b[i]) for i in range(a.shape[0])]) ## we interpret the first dimension as a timeseries multiplication
    else:
        res = np.matmul(a, b)
    if len(res.shape)>= 1 and res.shape[0] == 1:
        res = res[0]
    elif len(res.shape) >= 2 and res.shape[1] == 1:
        res = res[:, 0]
    if len(res.shape)>= 1 and res.shape[0] == 1:
        res = res[0]
    return res
    
def _as_list(value):
    if value is None:
        return []
    elif is_num(value):
        return [value]
    elif is_list(value) and is_nums(value):
        return [np.array(value)]
    elif is_pd(value):
        return [value]
    else:
        return as_list(value)

@pd2np
@apply_along_first_axis(base_shape = 2)
@mask_nans
def _matmul(matrix, rhs = None, lhs = None):
    matrices = _as_list(lhs) + _as_list(matrix) + _as_list(rhs)
    res = reducer(_mult, matrices)
    return res

def matmul(matrix, rhs = None, lhs = None, data = None, index = None, columns = None, power = 1):
    """

    :Example: simple matrix multiplication
    ----------
    >>> from pyg import *
    >>> rhs = np.array([1,2,3])
    >>> lhs = np.array([1,2,3])
    >>> matrix = np.array([[1,2,3],[4,5,6], [7,8,9]])    

    >>> assert matmul([1,2,3], [4,5,6]) == 32
    >>> assert eq(matmul(rhs,2), np.array([2, 4, 6]))
    >>> assert matmul(rhs,lhs) == 14
    >>> assert eq(matmul(matrix,rhs), np.array([14, 32, 50]))
    >>> assert matmul(matrix, rhs, lhs) == 228
    

    :Example: timeseries multiplication
    ----------
    >>> m = np.array([matrix] * 10)    
    >>> r = np.array([rhs] * 10)
    >>> l = np.array([lhs] * 10)
    
    >>> assert eq(matmul(r, l), np.array([14]*10))
    >>> assert eq(matmul(m, r, l), np.array([228]*10))
    >>> assert eq(matmul(matrix, r, l), np.array([228]*10))
    >>> assert eq(matmul(m, rhs, lhs), np.array([228]*10))
    >>> assert eq(matmul(m, r, lhs), np.array([228]*10))
    >>> assert eq(matmul(m, r), np.array([[14, 32, 50]] * 10))
    >>> assert eq(matmul(m, rhs), np.array([[14, 32, 50]] * 10))


    :Example: handling nans
    -----------------------
    from numpy import nan
    >>> matrix = np.array([[1,2,nan], [3,4,nan], [nan, nan, nan]])
    >>> rhs = np.array([1,2,np.nan])
    >>> assert eq(matmul(matrix, rhs), np.array([ 5., 11., nan]))
    >>> assert eq(matmul(matrix, rhs, rhs), 27)
    >>> assert matmul(rhs, lhs) == 5
    """    
    lhs_    = _as_list(lhs) 
    matrix_ = _as_list(matrix) 
    rhs_    = _as_list(rhs)
    pds = [m for m in lhs_ + matrix_ + rhs_ if is_pd(m)]
    lhs_ = [m.values if is_pd(m) else m for m in lhs_]
    rhs_ = [m.values if is_pd(m) else m for m in rhs_]
    if index is None:
        pd0 = [m for m in pds if m.shape[0] == matrix.shape[0]]
        if len(pd0):
            index = pd0[0].index
    if columns is None:
        pd1 = [m for m in pds if len(m.shape) == 2 and m.shape[1] == matrix.shape[1]]
        if len(pd1):
            columns = pd1[0].columns        
    res = _matmul(matrix = matrix, rhs = rhs_, lhs = lhs_)
    if power!=1:
        if power<0:
            res = v2na(res)
        res = res ** power
    if len(res.shape) == 1 and index is not None:
        res = pd.Series(res, index)
    elif len(res.shape) == 2 and (index is not None or columns is not None):
        res = pd.DataFrame(res, index, columns)        
    return res

# -*- coding: utf-8 -*-




from pyg_base import getargspec, getcallarg, first, Dict, loop, zipper, as_list, getargs, wrapper, skip_if_data_pd, dt, is_pd
from numba import njit
import numpy as np

__all__ = ['compiled']

def compiled(function):
    res = njit(nogil = True)(function)
    res.fullargspec = getargspec(function)
    return res

#from pyg.base import passthru; compiled = passthru ## use this to get valid code-coverage for compiled functions

first_ = loop(dict, list)(first)
second_ = loop(dict, list)(lambda value: value[1])


def _data_state(keys, values, output = 'data'):
    if isinstance(values, dict):
        return type(values)({k: _data_state(keys, v, output) for k, v in values.items()})
    elif isinstance(values, list):
        return type(values)([_data_state(keys, v, output) for v in values])
    output = as_list(output)
    assert keys[:len(output)] == output
    res = Dict(zip(output,values))
    if len(keys) > len(output):
        res['state'] = Dict(zipper(keys[len(output):], values[len(output):]))
    return res


def _ignore(ignore, data, state):
    state = {} if state is None else state
    if ignore is True:
        return None, {}
    elif ignore is False:
        return data, state
    ignore = as_list(ignore)
    if 'data' in ignore:
        data = None
    if 'state' in ignore:
        state = {}
    return data, state



@loop(dict, tuple, list)
def _mask(arg, mask, apply_axis):
    """
    removes nans based on a 1-d mask

    """
    res = arg
    f = len(mask)
    if isinstance(arg, np.ndarray):
        if (len(arg.shape) == 1 or apply_axis == 0) and arg.shape[0] == f:
            res = res[mask]
        elif len(arg.shape) > 1 and apply_axis == 1 and arg.shape[1] == f:
            res = res[:, mask]
        elif len(arg.shape) > 2 and apply_axis == 2 and arg.shape[2] == f:
            res = res[:, mask]
    return res

def _unmask1(arg, mask, apply_axis):
    n = len(mask[mask]) # len of no nans
    f = len(mask) # full length
    res = arg
    if (len(arg.shape) > 0 and apply_axis == 0) and arg.shape[0] == n:
        res = np.full( (f,) + arg.shape[1:], np.nan)
        res[mask] = arg
    elif len(arg.shape) > 1 and apply_axis == 1 and arg.shape[1] == n:
        res = np.full((arg.shape[0], f) + arg.shape[2:], np.nan)
        res[:, mask] = arg
    elif len(arg.shape) > 2 and apply_axis == 2 and arg.shape[2] == n:
        res = np.full((arg.shape[0], arg.shape[1], f) + arg.shape[3:], np.nan)
        res[:, mask] = arg
    return res
    

@loop(dict, tuple, list)
def _unmask(arg, masks):
    """
    returns nans based on a 1-d mask
    
    :Example:
    ----------
    >>> mask0 = np.array([True, False, True, True, False, True])
    >>> mask1 = np.array([True, False, True, True])
    >>> mask2 = np.array([True, True, False, True])
    >>> arg = np.array([[1,2,3.], [4,5,6.], [7,8,9.], [10,11,12]])
    >>> arg_ = _unmask1(_unmask1(arg, mask0, 0), mask1, 1)
    >>> arg1 = np.array([[1,2,3.], [4,5,6.], [7,8,9.]])
    >>> arg1_ = _unmask1(_unmask1(arg1, mask1, 0), mask1, 1)
    >>> arg2_ = _unmask1(_unmask1(arg1, mask2, 0), mask1, 1)
    
    
    >>> mask_nan_in_array(np.sum)(arg_, axis = 0)
    >>> mask_nan_in_array(np.sum)(arg, axis = 0)
    _unmask(mask_nan_in_array(np.sum)(arg, axis = 0), mask1, 1)
    
    >>> arg = _mask(_mask(arg, mask, apply_axis = 0), mask, 1)
    
    

    """
    res = arg
    if isinstance(arg, np.ndarray):
        if len(arg.shape) == 1: ## we give priorities to earlier dimensions
            masks = {apply_axis: mask for apply_axis, mask in masks.items() if len(mask[mask]) == arg.shape[0]}
            if len(masks) == 0:
                return res
            elif len(masks) > 1:
                unique_masks = set([tuple(mask) for mask in masks.values()])
                if len(unique_masks) > 1:
                    raise ValueError('There are multiple dimension possible to unmask the result %s'%masks)
            else:
                for apply_axis, mask in masks.items():
                    f = len(mask)
                    res = np.full(f, np.nan)
                    res[mask] = arg
                    return res
        elif len(arg.shape) == 2 and len(masks) and max(masks.keys()) == 3:
            raise ValueError('not implemented')
        for apply_axis, mask in masks.items():
            res = _unmask1(res, mask = mask, apply_axis = apply_axis)
    return res

    
class mask_nans(wrapper):
    """
    This wrapper allows us to operate in as if nan's are not actually provided
    """
    def __init__(self, function = None, apply_axis = None, exclude_any_nan = False):
        return super(mask_nans, self).__init__(function = function, apply_axis = apply_axis, exclude_any_nan = exclude_any_nan)

    def wrapped(self, *args, **kwargs):
        arg = getcallarg(self.function, args, kwargs)
        if not isinstance(arg, np.ndarray):
            return self.function(*args, **kwargs)        
        masks = {}
        mask = ~np.isnan(arg)
        args_, kwargs_ = args, kwargs
        for apply_axis in range(len(arg.shape)):
            if self.apply_axis is None or apply_axis in as_list(self.apply_axis):
                msk = mask
                for other_axis in range(len(arg.shape)):
                    if other_axis!=apply_axis:
                        msk = msk.min(axis = other_axis) if self.exclude_any_nan else msk.max(axis = other_axis)
                if not msk.min():
                    masks[apply_axis] = msk
                    args_ = tuple([_mask(a, msk, apply_axis) for a in args_])
                    kwargs_ = {k:_mask(v, msk, apply_axis) for k,v in kwargs_.items()}
                    arg = getcallarg(self.function, args_, kwargs_)
        res = self.function(*args_, **kwargs_)
        res = _unmask(res, masks)
        return res

@loop(dict, list, tuple)
def _stretch_over_time(arg, t):
    if isinstance(arg, np.ndarray) and len(arg.shape) and arg.shape[0] == t:
        return arg
    elif is_pd(arg) and len(arg) == t:
        return arg
    else:
        return np.array([arg] * t)

@loop(dict, list, tuple)
def _i(value, i, t = None):
    if t is not None:
        if hasattr(value, '__len__') and len(value) == t:    
            return value.iloc[i] if is_pd(value) else value[i]
        else:
            return value
    else:
        if hasattr(value, '__len__'):    
            return value.iloc[i] if is_pd(value) else value[i]
        else:
            return value

class apply_along_first_axis(wrapper):
    """
    applies a function along 1st axis (similar to np.apply_along_axis)
    
    :Parameters:
    ------------
    base_shape:
        defines the minimum shape of the non-time axis object.

    state:
        allows the function to take the output of the previous row and feed it back as input to next row calculation
    
    
    :Example: base_shape
    ---------
    >>> vector = np.array([1,2,3])
    >>> mtrx = np.array([[1,2], [3,4], [5,6]])
    >>> assert mtrx.shape == (3,2)

    >>> f1 = apply_along_first_axis(np.sum, base_shape = 1) ## I EXPECT to operate on rows, so if you see a matrix, apply
    >>> assert eq(f1(mtrx), np.array([ 3,  7, 11]))
    >>> assert eq(f1(vector), 6)
        
    >>> f2 = apply_along_first_axis(np.sum, base_shape = 2) ## I am EXPECTING a matrix in each time unit, so a isn't time at all, it is a single valued matrix 
    >>> assert f2(mtrx) == np.sum(mtrx)
    >>> assert f2(vector) == np.sum(vector)
    
    :Example: state
    ------------------
    >>> f = lambda vector, running_total = 0: np.sum(vector) + running_total
    >>> cumsum = apply_along_first_axis(f, base_shape = 1, state = 'running_total')
    >>> assert eq(cumsum(mtrx), np.array([3,10,21]))
    """
    def __init__(self, function = None, base_shape = 1, state = None, message = None):
        return super(apply_along_first_axis, self).__init__(function = function, base_shape = base_shape, state = state, message = message)

    def wrapped(self, *args, **kwargs):
        arg = getcallarg(self.function, args, kwargs)
        if not isinstance(arg, np.ndarray) or len(arg.shape)<=self.base_shape:
            return self.function(*args, **kwargs)        
        t = arg.shape[0]
        args_, kwargs_ = args, kwargs # _stretch_over_time((args, kwargs), t = t)       
        if self.state is None:
            res = [self.function(*_i(args_,i,t), **_i(kwargs,i,t)) for i in range(t)]
        else:
            i = 0
            t0 = dt()
            states = as_list(self.state)
            res = [self.function(*_i(args_,i,t), **_i(kwargs,i,t))]
            for i in range(1, t):
                kw = _i(kwargs_, i, t)
                for state in states:
                    kw[state] = res[i-1][state] if isinstance(res[i-1], dict) else res[i-1]
                res.append(self.function(*_i(args_,i,t), **kw))
                if self.message and i % self.message == 0:
                    t1 = dt()
                    print(t1, 'calculated %i of %i in %s'%(i, t, t1 - t0))
                    t0 = t1
        if len(res) and isinstance(res[0], dict):
            return type(res[0])({k : np.array([r.get(k) for r in res]) for k in res[0].keys()})
        else:
            return np.array(res)

import pandas as pd
from pyg.base import wrapper, is_ts, is_dict, is_tuple, getargspec, loop, first, Dict, zipper

class persist_data(wrapper):
    """
    We work with state-persisting function that return a variable called 'data'
    
    :Example:
    -------
    >>> from pyg import *
    >>> ts = pd.Series(np.random.normal(0,1,(1000)), drange(-999))
    >>> ts[ts<0.1] = np.nan

    We can run ts in one go:
        
    >>> both = ffill_(ts)

    Or split it into two parts:

    >>> old = ffill_(ts.iloc[:500])
    
    old.data is the data for first 500 entries. We can use this data to speed up later calculations...
    
    >>> new = ffill_(ts.iloc[500:], **(old-'data')) 
    
    We note: 
    1) the 'data' variable must be remove
    2) We need to glue together old.data and new.data
    
    However... persist_data does the work for you:
        
    >>> persist_ffill_ = persist_data(ffill_)
    >>> glued_with_old = persist_ffill_(ts, **old)  # Note that we pass on data and the full ts, but calculation is done only on the 'new' bit of ts
    >>> glued_just_new = persist_ffill_(ts.iloc[500:], **old)  

    >>> assert eq(glued_with_old, both)    
    >>> assert eq(glued_just_new, both)    
    
    """
    def persist_outputs(self, *args, **kwargs):
        output = as_list(getattr(self.function,'output', 'data'))
        old = {o: kwargs.pop(o) for o in output}
        old_ts = {k:v for k, v in old.items() if is_ts(v) and len(v)}
        if len(old_ts) == 0:
            return self.function(*args, **kwargs, **old)
        old_t1 = set([v.index[-1] for v in old_ts.values()])
        if len(old_t1) > 1:
            raise ValueError(f'multiple timeseries and multiple endates {old_t1}')
        cutoff = list(old_t1)[0]
        args_ = [v[v.index>cutoff] if is_ts(v) and len(v) else v for v in args]
        kwargs_ = {k : v[v.index>cutoff] if is_ts(v) and len(v) else v for k, v in kwargs.items()}
        res = self.function(*args_, **kwargs_, **old)
        

            
    
    def wrapped(self, *args, **kwargs):
        data = kwargs.pop('data', None)
        if data is not None and len(data)>0 and is_ts(data):
            cutoff = data.index[-1]
            args_ = [v[v.index>cutoff] if is_ts(v) and len(v) else v for v in args]
            kwargs_ = {k : v[v.index>cutoff] if is_ts(v) and len(v) else v for k, v in kwargs.items()}
            res = self.function(*args_, **kwargs_)
            if res is None:
                return data
            if is_ts(res):
                new = res
            elif is_dict(res):
                new = res.get('data')
            elif is_tuple(res):
                new = res[0]
            else:
                raise ValueError('result of persistence data must be a timeseries or a dict \n%s '%res)
            if new is None:
                both = data
            elif is_ts(new):
                both = data if len(new) == 0 else pd.concat([data[data.index<new.index[0]], new])
            else:
                raise ValueError('data must be None or a timeseries')                
            if is_ts(res):
                return both
            elif is_dict(res):
                res['data'] = both
                return res
            elif is_tuple(res):
                return (both,) + res[1:]
        else:
            return self.function(*args, **kwargs)




from pyg_timeseries._rolling import _rolling_window
from pyg_timeseries._decorators import first_, _data_state

__all__ = ['rolling_median', 'rolling_median_', 'expanding_median']

def rolling_median(a, n, axis = 0, data = None, state = None):
    """
    equivalent to pandas a.rolling(n).median().
    
    - works with np.arrays
    - handles nan without forward filling.
    - supports state parameters
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
    n: int
        size of rolling window
    axis : int, optional
        0/1/-1. The default is 0.    
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
        
    :Example: agreement with pandas
    --------------------------------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> panda = a.rolling(10).median(); ts = rolling_median(a,10)
    >>> assert abs(ts-panda).max()<1e-10   
    >>> ab = pd.DataFrame(dict(a = np.random.normal(0,1,10000), b = np.random.normal(0,1,10000)), drange(-9999))
    >>> panda = ab.rolling(10).median(); ts = rolling_median(ab,10)
    >>> assert abs(ts-panda).max().max()<1e-10   

    :Example: nan handling
    ----------------------
    Unlike pandas, timeseries does not include the nans in the rolling calculation: it skips them.
    Since pandas rolling engine does not skip nans, they propagate. 
    In fact, having removed half the data points, rolling(10) will return 99% of nans

    >>> a[a<0.1] = np.nan
    >>> panda = a.rolling(10).median(); ts = rolling_median(a,10)
    >>> print('#original:', len(nona(a)), 'timeseries:', len(nona(ts)), 'panda:', len(nona(panda)), 'data points')
    #original: 4634 timeseries: 4625 panda: 4 data points

    :Example: state management
    --------------------------
    One can split the calculation and run old and new data separately.

    >>> old = a.iloc[:5000]        
    >>> new = a.iloc[5000:]    
    >>> ts = rolling_median(a,10)
    >>> old_ts = rolling_median_(old,10)
    >>> new_ts = rolling_median(new, 10, **old_ts)    
    >>> assert eq(new_ts, ts.iloc[5000:])

    :Example: dict/list inputs
    ---------------------------
    >>> assert eq(rolling_median(dict(x = a, y = a**2),10), dict(x = rolling_median(a,10), y = rolling_median(a**2,10)))
    >>> assert eq(rolling_median([a,a**2],10), [rolling_median(a,10), rolling_median(a**2,10)])

    """
    import bottleneck as bn
    state = state or dict(vec = None)
    return first_(_rolling_window(a,n,n, func = bn.move_median, axis = axis, **state))

def rolling_median_(a, n, axis = 0, data = None, instate = None):
    """
    Equivalent to rolling_median(a) but returns also the state. 
    For full documentation, look at rolling_median.__doc__    
    """
    import bottleneck as bn
    state = instate or dict(vec = None)
    return _data_state(['data','vec'],_rolling_window(a, window = n, min_count = n, func = bn.move_median, axis = axis, **state))

rolling_median_.output = ['data','state']


def expanding_median(a, axis = 0):
    """
    equivalent to pandas a.expanding().median().
    
    - works with np.arrays
    - handles nan without forward filling.
    - There is no state-aware version since this requires essentially the whole history to be stored.
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
        
    axis : int, optional
        0/1/-1. The default is 0.
            
    :Example: agreement with pandas
    --------------------------------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> panda = a.expanding().median(); ts = expanding_median(a)
    >>> assert eq(ts,panda)    

    :Example: nan handling
    ----------------------
    Unlike pandas, timeseries does not forward fill the nans.    

    >>> a[a<0.1] = np.nan
    >>> panda = a.expanding().median(); ts = expanding_median(a)
    
    >>> pd.concat([panda,ts], axis=1)
    >>>                    0         1
    >>> 1993-09-23  1.562960  1.562960
    >>> 1993-09-24  0.908910  0.908910
    >>> 1993-09-25  0.846817  0.846817
    >>> 1993-09-26  0.821423  0.821423
    >>> 1993-09-27  0.821423       NaN
    >>>              ...       ...
    >>> 2021-02-03  0.870358  0.870358
    >>> 2021-02-04  0.870358       NaN
    >>> 2021-02-05  0.870358       NaN
    >>> 2021-02-06  0.870358       NaN
    >>> 2021-02-07  0.870353  0.870353
    
    :Example: dict/list inputs
    ---------------------------
    >>> assert eq(expanding_median(dict(x = a, y = a**2)), dict(x = expanding_median(a), y = expanding_median(a**2)))
    >>> assert eq(expanding_median([a,a**2]), [expanding_median(a), expanding_median(a**2)])

    """

    import bottleneck as bn
    return first_(_rolling_window(a,0,1, vec = None, func = bn.move_median, axis = axis))



from pyg_timeseries._rolling import _rolling_window
from pyg_base import pd2np, loop_all
from pyg_timeseries._decorators import first_, compiled, _data_state
import numpy as np

__all__ = ['rolling_min', 'rolling_min_', 'expanding_min', 'expanding_min_']

def rolling_min(a, n, axis = 0, data = None, state = None):
    """
    equivalent to pandas a.rolling(n).min().
    
    - works with np.arrays
    - handles nan without forward filling.
    - supports state parameters
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
    n: int
        size of rolling window
    axis : int, optional
        0/1/-1. The default is 0.    
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
        
    :Example: agreement with pandas
    --------------------------------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> panda = a.rolling(10).min(); ts = rolling_min(a,10)
    >>> assert abs(ts-panda).min()<1e-10   

    :Example: nan handling
    ----------------------
    Unlike pandas, timeseries does not include the nans in the rolling calculation: it skips them.
    Since pandas rolling engine does not skip nans, they propagate. 
    In fact, having removed half the data points, rolling(10) will return 99% of nans

    >>> a[a<0.1] = np.nan
    >>> panda = a.rolling(10).min(); ts = rolling_min(a,10)
    >>> print('#original:', len(nona(a)), 'timeseries:', len(nona(ts)), 'panda:', len(nona(panda)), 'data points')
    >>> #original: 4534 timeseries: 4525 panda: 6 data points

    :Example: state management
    --------------------------
    One can split the calculation and run old and new data separately.

    >>> old = a.iloc[:5000]        
    >>> new = a.iloc[5000:]    
    >>> ts = rolling_min(a,10)
    >>> old_ts = rolling_min_(old,10)
    >>> new_ts = rolling_min(new, 10, **old_ts)    
    >>> assert eq(new_ts, ts.iloc[5000:])

    :Example: dict/list inputs
    ---------------------------
    >>> assert eq(rolling_min(dict(x = a, y = a**2),10), dict(x = rolling_min(a,10), y = rolling_min(a**2,10)))
    >>> assert eq(rolling_min([a,a**2],10), [rolling_min(a,10), rolling_min(a**2,10)])

    """
    import bottleneck as bn
    state = state or {}
    return first_(_rolling_window(a,n,n, func = bn.move_min, axis = axis, **state))

def rolling_min_(a, n, vec = None, axis = 0, data = None, instate = None):
    """
    Equivalent to rolling_min(a) but returns also the state. 
    For full documentation, look at rolling_min.__doc__    
    """
    import bottleneck as bn
    state = instate or {}
    return _data_state(['data','vec'],_rolling_window(a,n,n, func = bn.move_min, axis = axis, **state))

rolling_min_.output = ['data','state']

@loop_all
@pd2np
@compiled
def _expanding_min(a, m = np.nan):
    res = np.empty_like(a)
    for i0 in range(a.shape[0]):
        if np.isnan(a[i0]):
            res[i0] = np.nan
        else:
            res[i0] = m = m if m<a[i0] else a[i0]
            break
    for i in range(i0+1, a.shape[0]):
        if np.isnan(a[i]):
            res[i] = np.nan
        else:
            res[i] = m = m if m<a[i] else a[i]
    return res, m


def expanding_min_(a, axis = 0, data = None, instate = None):
    """
    Equivalent to a.expanding().min() but returns the full state: i.e. both 
    data: the expanding().min()
    m: the current minimum
    """
    state = instate or {}
    return _data_state(['data','m'], _expanding_min(a, axis = axis, **state))

expanding_min_.output = ['data','state']


def expanding_min(a, axis = 0, data = None, state = None):
    """
    equivalent to pandas a.expanding().min().
    
    - works with np.arrays
    - handles nan without forward filling.
    - supports state parameters
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
        
    axis : int, optional
        0/1/-1. The default is 0.
    
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
        
    :Example: agreement with pandas
    --------------------------------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> panda = a.expanding().min(); ts = expanding_min(a)
    >>> assert eq(ts,panda)    

    :Example: nan handling
    ----------------------
    Unlike pandas, timeseries does not forward fill the nans.    

    >>> a[a<0.1] = np.nan
    >>> panda = a.expanding().min(); ts = expanding_min(a)
    
    >>> pd.concat([panda,ts], axis=1)
    >>>                    0         1
    >>> 1993-09-24       NaN       NaN
    >>> 1993-09-25       NaN       NaN
    >>> 1993-09-26  0.775176  0.775176
    >>> 1993-09-27  0.691942  0.691942
    >>> 1993-09-28  0.691942       NaN
    >>>              ...       ...
    >>> 2021-02-04  0.100099  0.100099
    >>> 2021-02-05  0.100099       NaN
    >>> 2021-02-06  0.100099       NaN
    >>> 2021-02-07  0.100099  0.100099
    >>> 2021-02-08  0.100099  0.100099
    
    :Example: state management
    --------------------------
    One can split the calculation and run old and new data separately.
    
    >>> old = a.iloc[:5000]        
    >>> new = a.iloc[5000:]    
    >>> ts = expanding_min(a)
    >>> old_ts = expanding_min_(old)
    >>> new_ts = expanding_min(new, **old_ts)    
    >>> assert eq(new_ts, ts.iloc[5000:])

    :Example: dict/list inputs
    ---------------------------
    >>> assert eq(expanding_min(dict(x = a, y = a**2)), dict(x = expanding_min(a), y = expanding_min(a**2)))
    >>> assert eq(expanding_min([a,a**2]), [expanding_min(a), expanding_min(a**2)])

    """
    state = state or {}
    return first_(_expanding_min(a, axis = axis, **state))




import numpy as np
import pandas as pd
from scipy.optimize import minimize
TOLERANCE = 1e-10
from pyg_base import pd2np, is_df, skip_if_data_pd_or_np, try_nan, as_list, reducer, is_num, is_nums, is_pd, is_tss, is_list, df_concat, df_reindex
from pyg_timeseries._rolling import v2na, na2v, ffill
from pyg_timeseries._decorators import mask_nans, apply_along_first_axis


def _risk_budget_objective_error(weights, args):

    # The covariance matrix occupies the first position in the variable
    covariances = args[0]

    # The desired contribution of each asset to the portfolio risk occupies the
    # second position
    assets_risk_budget = args[1]

    # We convert the weights to a matrix
    weights = np.matrix(weights)

    # We calculate the risk of the weights distribution

    gradient = covariances * weights.T
    portfolio_risk = np.sqrt(weights * gradient)[0,0]

    # portfolio_risk = _allocation_risk(weights, covariances)

    # We calculate the contribution of each asset to the risk of the weights
    # distribution
    assets_risk_contribution = np.multiply(weights.T, gradient) / portfolio_risk

    # We calculate the desired contribution of each asset to the risk of the
    # weights distribution
    assets_risk_target = np.asmatrix(np.multiply(portfolio_risk, assets_risk_budget))

    # Error between the desired contribution and the calculated contribution of
    # each asset
    error = sum(np.square(assets_risk_contribution - assets_risk_target.T))[0, 0]

    # It returns the calculated error
    return error


def _risk_budget_objective_error_no_gradient(weights, args):
    covariances = args[0]
    n = covariances.shape[0]  ** 2
    assets_risk_budget = args[1]
    w = np.matrix(weights).T ## 1 x n matrix
    gradient = covariances * w ## n x 1
    assets_risk_contribution = np.multiply(w , gradient)
    total = np.sum(assets_risk_contribution)
    assets_risk_contribution /= total
    difference = assets_risk_contribution.T - assets_risk_budget
    error = n * np.sum(np.square(difference))
    return error


def _risk_budget_objective_error_with_gradient(weights, args):
    covariances = args[0]
    n = covariances.shape[0]  ** 2
    assets_risk_budget = args[1]
    w = np.matrix(weights) ## 1 x n matrix
    gradient = 0.5 * np.array(covariances * w.T)[:,0] ## n x 1
    assets_risk_contribution = weights * gradient
    total = np.sum(assets_risk_contribution)
    assets_risk_contribution /= total
    difference = assets_risk_contribution - assets_risk_budget
    error = n * np.sum(np.square(difference))
    derror_by_dweights = n * (difference * assets_risk_contribution) * (gradient / total)
    return error, derror_by_dweights


_constraints = ({'type': 'eq', 'fun': lambda x: np.sum(x) - 1.0},
                   {'type': 'ineq', 'fun': lambda x: x})


@pd2np
@apply_along_first_axis(base_shape = 2, state = 'weights', message = 100)
def _riskparity(covariances, assets_risk_budget = None, weights = None, tol = None, method = 'pyrb'):

    # Restrictions to consider in the optimisation: only long positions whose
    # sum equals 100%

    if assets_risk_budget is None:
        n = covariances.shape[0]
        arb = np.full(n, 1. / n)
    else:
        arb = assets_risk_budget.copy()
        arb[np.isnan(arb)] = 0

    d = np.diagonal(covariances)
    msk = ~np.maximum(np.isnan(d), d <=0, arb == 0)

    masked_c = covariances[msk][:,msk]
    masked_a = arb[msk]
    masked_w = None if weights is None else weights[msk]

    fit = np.full_like(arb, np.nan)
    n = masked_c.shape[0]
    if n == 0:
        return fit
    
    if method == 'pyrb':
        try:
            import pyrb
        except Exception:
            raise ValueError('need to install pyrb from https://github.com/jcrichard/pyrb \npip install git+https://github.com/jcrichard/pyrb \nTo')
        try:
            RB = pyrb.RiskBudgeting(masked_c, masked_a)
            RB.solve()
            fit[msk] = RB.x
        except ValueError:
            pass
        return fit #, matmul(masked_c, RB.x, RB.x)

    if masked_w is None:
        x0 = np.full(n,  1./n)
    else:
        x0_mask = np.isnan(masked_w)
        if x0_mask.min():
            x0 = np.full(n, 1./n)
        else:
            x0 = masked_w
            x0[x0_mask] = np.mean(x0[~x0_mask])
    x0 = x0 / np.sum(x0)
    tol = tol or TOLERANCE
    
    # Optimisation process in scipy
    if method == 'old':
        optimize_result = minimize(fun = _risk_budget_objective_error,
                                    x0 = x0,
                                    args = [covariances, arb],
                                    method = 'SLSQP',
                                    constraints = _constraints,
                                    tol = tol,
                                    options= {'disp': False})
    elif method == 'gradient':        
        optimize_result = minimize(fun = _risk_budget_objective_error_with_gradient,
                                    x0 = x0,
                                    args = [covariances, arb],
                                    method = 'SLSQP',
                                    jac = True,
                                    constraints = _constraints,
                                    tol = tol,
                                    options= {'disp': False})
    elif method == 'sum_to_one':
        optimize_result = minimize(fun = _risk_budget_objective_error_no_gradient,
                                    x0 = x0,
                                    args = [covariances, arb],
                                    method = 'SLSQP',
                                    constraints = _constraints,
                                    tol = tol,
                                    options= {'disp': False})
    # Recover the weights from the optimised object
    fit[msk] = optimize_result.x
    return fit #, matmul(masked_c, optimize_result.x, optimize_result.x)

def _data_ok(data, covariances, columns):
    if data is None or len(data.shape)!=2: ## shape OK
        return False
    if data.shape[1] != covariances.shape[1]: ## matching covariances
        return False
    if is_df(data) and columns is not None and list(columns)!=list(data.columns): # match columns if available
        return False
    return True
        

def riskparity(covariances, assets_risk_budget = None, columns = None, index = None, data = None, tol = None, method = 'pyrb'):
    """
    Designed to take the output from the ewmcovar calculation or a simple covariances matrix.
    Calculates risk parity weights given covariances and target budgets.
    If data is provided, will skip calculations for previously done optimizations.
    
    The faster implementation is using pyrb available from https://github.com/jcrichard/pyrb
    This relies on quadprog https://github.com/quadprog/quadprog which in turns requires a C compiler like the MS one.
    
    The slower implementation is using scipy:
    https://thequantmba.wordpress.com/2016/12/14/risk-parityrisk-budgeting-portfolio-in-python/
    or 
    https://quantdare.com/risk-parity-in-python/ 
    
    :Example:
    ---------
    >>> covariances = np.array([ [1,0.8,0.1], [0.8,1,0.1], [0.1, 0.1, 1] ]) 
    >>> weights = riskparity(covariances)
    >>> print(weights) 

    :Example: np.nans
    ------------------
    >>> covariances = np.array([ [1,0.8,0.1, np.nan], [0.8,1,0.1,np.nan], [0.1, 0.1, 1, np.nan] , [np.nan] * 4]) 
    >>> weights = riskparity(covariances)
    >>> print(weights) 

    :Parameters:
    ------------
    covariances : matrix
        covariance matrix or a timeseries of covariances
    assets_risk_budget : np.array, optional
        target risk budget per asset
    weights : np.array, optional
        initial weights for estimated allocation. The default is None.
    tol : float, optional
        Calculation tolerance. The default is None.
    columns : list, optional
        columns names. The default is None.
    index : list, optional
        index dates. The default is None.

    :Returns:
    -------
    weights : np.array
        The risk parity weights
        
        
    :Example:
    ---------
    >>> from pyg import *
    >>> rs = dictable(ticker = ['ESA Index', 'TYA Comdty', 'CLA Comdty', 'ECA Curncy'])
    >>> rs = rs(vrtn = lambda ticker: get_data('ticker', 'bbg', item = 'vrtn', ticker = ticker))
    >>> rs = rs(vrtn = lambda vrtn: vrtn[0] if is_df(vrtn) else vrtn)
    >>> vrtn = df_concat(rs.vrtn, rs.ticker)
    >>> vrtn1w = v2na(vrtn.resample('w').sum())


    >>> cor = ewmcorr_(vrtn1w, 52)['data']
    >>> cov = ewmcovar_(vrtn1w, 52)
    >>> res = riskparity(cor['data']**2, columns = cor['columns'], index = cor['index'])
    >>> res[dt(1985):].plot()
    >>> res1 = riskparity(cov['data'], columns = cov['columns'], index = cov['index'])
    >>> res[dt(1985):].plot()
    >>> res1[dt(1985):].plot()

    args = (cov['data'],)
    kwargs = dict(index = cov['index'], columns = cov['columns'])
    self = _eigenvalues
    >>> eigenvalues(cov['data'], index = cov['index']).plot()
    >>> eigenvalues(cor['data'], index = cor['index']).plot()
    >>> eigenvectors(cor['data'], n = 3, columns = cor['columns'], index = cor['index'], sort = -1)

    :Example: repeating the calculation, this time 
    ---------
    >>> res2 = riskparity(cov['data'], columns = cov['columns'], index = cov['index'], data = res1.iloc[:1800])
    >>> ratio = res2/res1
    >>> assert ratio.max().max() < 1.0001 and ratio.min().min() > 0.9999
    """
    arb = assets_risk_budget; cov = covariances; idx = index; weights = None
    cov = ffill(cov)
    if _data_ok(data, covariances, columns):      
        if is_df(data) and index is not None: ## previous result is a timeseries
            data = data.iloc[:-1] ## remove last calculation
            mask = ~np.array([date in data.index for date in index])
            cov = cov[mask]
            idx = index[mask]
            weights = data.iloc[-1].values
            if is_df(arb):
                arb = arb[mask]
        elif not is_df(data):
            data = data[:-1]
            n = len(data)
            cov = cov[n:]
            if arb is not None and len(arb.shape) == 2:
                arb = arb[n:]
            if idx is not None:
                idx = idx[n:]
            weights = data[-1]
    else:
        data = None

    if isinstance(arb, list):
        if is_nums(arb):
            arb = np.array(arb)
        elif is_tss(arb):
            arb = v2na(df_reindex(df_concat(arb), index, method = 'ffill'))
    if is_pd(arb):
        arb = arb.values
        if len(arb.shape) == 2:
            arb = ffill(arb)
        arb = na2v(arb)
        
    res = _riskparity(covariances = cov, assets_risk_budget = arb, weights = weights, tol = tol, method = method)
    
    if len(res.shape) == 1:
        return res if index is None else pd.Series(res, idx)
    else:
        res = ffill(res)
        if columns is not None or index is not None:
            res = pd.DataFrame(data = res, columns = columns, index = idx) 
            if data is not None:
                res = pd.concat([data, res]).sort_index()
            return res
        else:
            if data is not None:
                res = np.concatenate([data, res])
            return res

    # -*- coding: utf-8 -*-




from pyg_timeseries._rolling import _vec, _rolling_window
from pyg_base import Dict
from pyg_timeseries._decorators import first_, _data_state

__all__ = ['rolling_rank', 'rolling_rank_', 'expanding_rank']

def rolling_rank(a, n, axis = 0, data = None, state = None):
    """
    returns a rank of the current value within a given window, scaled to be -1 if it is the smallest and +1 if it is the largest
    - works on mumpy arrays too
    - skips nan, no ffill
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
        
    n: int
        window size 

    axis : int, optional
        0/1/-1. The default is 0.

    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
    
    :Example:
    -------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series([1.,2., np.nan, 0., 4., 2., 3., 1., 2.], drange(-8))
    >>> rank = rolling_rank(a, 3)
    >>> assert eq(rank.values, np.array([np.nan, np.nan, np.nan, -1, 1, 0, 0, -1, 0]))
    >>> # 0 is smallest in [1,2,0] so goes to -1
    >>> # 4 is largest in [2,0,4] so goes to +1
    >>> # 2 is middle of [0,4,2] so goes to 0
    
    :Example: numpy equivalent
    ------------------------------------------
    >>> assert eq(rolling_rank(a.values, 10), rolling_rank(a, 10).values)  
    
    :Example: state management
    --------------------------------------------
    >>> a = np.random.normal(0,1,10000)
    >>> old = rolling_rank_(a[:5000], 10) # grab both data and state
    >>> new = rolling_rank(a[5000:], 10, **old)
    >>> assert eq(np.concatenate([old.data,new]), rolling_rank(a, 10))

    """
    import bottleneck as bn
    state = state or {}
    return first_(_rolling_window(a,n,n, func = bn.move_rank, axis = axis, **state))

def rolling_rank_(a, n, axis = 0, data = None, instate = None):
    """
    Equivalent to rolling_rank(a) but returns also the state variables.
    For full documentation, look at rolling_rank.__doc__    
    """
    import bottleneck as bn
    state = instate or {}
    return _data_state(['data','vec'],_rolling_window(a,n,n,func = bn.move_rank, axis = axis, **state))

rolling_rank_.output = ['data','state']


def expanding_rank(a, axis = 0):
    """
    returns a rank of the current value within history, scaled to be -1 if it is the smallest and +1 if it is the largest
    - works on mumpy arrays too
    - skips nan, no ffill
    
    :Parameters:
    ------------
    a : array, pd.Series, pd.DataFrame or list/dict of these
        timeseries
        
    axis : int, optional
        0/1/-1. The default is 0.

    :Example:
    -------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series([1.,2., np.nan, 0.,4.,2.], drange(-5))
    >>> rank = expanding_rank(a)
    >>> assert eq(rank, pd.Series([0, 1, np.nan, -1, 1, 0.25], drange(-5)))
    >>> #
    >>> # 2 is largest in [1,2] so goes to 1; 
    >>> # 0 is smallest in [1,2,0] so goes to -1 etc.
    
    :Example: numpy equivalent
    ------------------------------------------
    >>> assert eq(expanding_rank(a.values), expanding_rank(a).values)  
    

    """
    import bottleneck as bn
    return first_(_rolling_window(a,0,1, vec = None, func = bn.move_rank, axis = axis))



import pandas as pd
import numpy as np
from pyg_base import pd2np, is_pd, loop, is_df, df_reindex, df_concat, is_tss
from pyg_timeseries._matmul import matmul
from pyg_timeseries._rolling import ffill, na2v
from pyg_timeseries._pandas import reindex_3d


### single pass position calculation

@loop(list, tuple)
@pd2np
def _as_int(pos):
    res = pos.copy()
    res[np.isnan(res)] = 0    
    res = res.astype(int) ## This is rounding down
    return res

@pd2np
def _int_nearest_target(target, init = None):
    """
    makes init goes within nearest unit of target size
    
    :Parameters:
    ============
    init: int array
        initial value
        
    target: float array
        target position to get to
        
    :Example:
    =========

    This ensures our initial position is never "above" the target position
    >>> target = np.array([0.2, 0.2, 0.2])
    >>> init = np.array([3,0,-3])
    >>> assert list(_int_nearest_target(target, init)) == [1,0,0]
    >>> assert list(_int_nearest_target(target)) == [0,0,0]
    """
    if init is None:
        init = np.zeros(target.shape)
    d = (target - init).astype(int)
    res = init + d
    return res

    

def _minimize_tracking_error(covariance, target, multiplier = None, position = None, relative_cost = None, min_change = 0.01, search = 2):
    n = target.shape[0]
    ns = np.arange(n)
    res = _int_nearest_target(target, position)    
    covariance[np.isnan(covariance)] = 0.0
    if is_pd(multiplier):
        multiplier = multiplier.values
    if multiplier is not None:
        multiplier[np.isnan(multiplier)] = 0.0
        rr = np.array(np.matmul(np.reshape(multiplier, (n,1)), np.reshape(multiplier, (1,n))))
        cr = np.array(covariance) * rr
    else:
        cr = np.array(covariance)
    j = 0
    m = -1
    moved = set()
    d = na2v(target - res) # if there is a nan in target, we do not trade
    gradient = matmul(cr, d)
    base = matmul(gradient, d)

    while m < -abs(min_change) and j < 20:
        m = 0
        sgn = np.sign(d)
        gbase =  -2 * gradient * sgn ## theoretical gradient
        mask = gbase < 0
        gbase = gbase[mask] # only trade if gradient is down
        nn = ns[mask]
        gn = sorted(zip(gbase, nn)) 
        ok = True
        found = 0
        dbase = [0] * len(ns)
        for g, i in gn: 
            ## we start from the lowest gradient and work our way up. Looking for valid candidates 
            ## The moment we find a enough candidates to perform a valid reduction in tracking error, we reduce. 
            ## if search = 0; we search exhaustively manually. search = 1 is potentially less accurate but faster. 
            ## The assumption being is that if there is another candidate, it will show in the next round of reduction anyway
            if ok and g < 0 and i not in moved:
                r = res.copy()
                r[i] = r[i] + sgn[i]
                d = na2v(target - r)
                dbase[i] = m = matmul(cr, d, d) / base - 1                
                if m < -abs(min_change):
                    found +=1
                    if search > 0 and found >= search:
                        ok = False
        if found > 1:
            m = min(dbase)
            i = dbase.index(m)
            
        if m < -abs(min_change):
            moved.add(i)
            res[i] = res[i] + sgn[i]
            d = na2v(target - res) # if there is a nan in target, we do not trade
            gradient = matmul(cr, d)
            base = matmul(gradient, d)    
            if base <= 0:
                return res, base
            j = j +1

    return res, base
                
@loop(list, tuple)
def _as_np(value):
    return value.values if is_pd(value) else value
 
@loop(list, tuple)
def _post_data(value, n, i):
    if isinstance(value, np.ndarray) and len(value) == n:
        return value[i:]
    else:
        return value

def _get_i(value, n, i):
    if isinstance(value, np.ndarray) and len(value) == n:
        return value[i]
    else:
        return value
        

def minimize_tracking_error(covariance, target, multiplier = None, min_change = 0.01, covariance_index = None, search = 2, data = None, errors = None):
    """
    Tracks a target position within an int, while minimizing the tracking error defined as 
    
    Tracking error = (data-target)^T x Covariance x (data-target)

    Parameters
    ----------
    covariance : 3d np.ndarray 
        covariance of the quote time data time x n x n 
    target : 2d pd.DataFrame or np.ndarray
        The position we want to take in each of the n assets, in FRACTIONS
    multiplier : 2d pd.DataFrame, optional
        Each asset may have a multiplier. This can be due to e.g. price being in a different currency or the quote (in yield) change translating into actual value changes via dv01 multiplier.
    min_change : float<1, optional
        We stop trying to optimize when we can improve the tracking error by at most min_change. The default is 0.01.
    data : pd.DataFrame, optional
        The previous output from this function. The default is None.
        The output tracks the target positions but is always an integer.

    Returns
    -------
    data : pd.DataFrame
        A position that tracks "target" but is:
            - always an integer
            - minimizes tracking error 

    """
    if is_tss(target):
        target = df_concat(target, method = 'ffill')
    if data is not None and len(data) == len(target):
        return dict(data = data, errors = errors)

    if is_df(target):
        multiplier = df_reindex(multiplier, target, method = 'ffill')
        if covariance_index is not None:
            covariance = reindex_3d(covariance, target.index, covariance_index, method = 'ffill')

    c, t, r, d, e = _as_np([covariance, target, multiplier, data, errors])
    c = na2v(ffill(c))
    t = na2v(ffill(t))
    if d is not None and len(d):        
        c, t, r = _post_data([c, t, r], len(c), len(d))
        position = d[-1]
    else:
        position = None
    n = len(c)
    positions = []
    tracking_errors = []
    for i in range(len(c)):
        if i % 100 == 0:
            print('running %i of %i'%(i, len(c)))
        position, tracking_error = _minimize_tracking_error(covariance = _get_i(value = c, n = n, i = i), 
                                                            target = _get_i(value = t, n = n, i = i), 
                                                            multiplier = _get_i(value = r, n = n, i = i), 
                                                            position = position, 
                                                            min_change = min_change,
                                                            search = search)
        positions.append(position)
        tracking_errors.append(tracking_error)
    positions = np.array(positions)
    tracking_errors = np.array(tracking_errors)
    if d is not None and len(d):
        positions = np.concatenate([d, positions])
    if e is not None and len(e):
        tracking_errors = np.concatenate([e, tracking_errors])
    if is_df(target):
        return dict(data = pd.DataFrame(positions, target.index, target.columns), errors = pd.Series(tracking_errors, target.index))
    else:
        return dict(data = positions, errors = tracking_errors)

minimize_tracking_error.output = ['data', 'errors']


from pyg_base import df_reindex, is_df, is_list, pd2np, loop, Dict, is_nums, is_pd, is_arr, is_num, df_concat, is_series, is_strs, dt, is_tss, is_ts
from pyg_timeseries._matmul import matmul
from pyg_timeseries._rolling import buffer, ffill, v2na
from pyg_timeseries._pandas import reindex_3d
import numpy as np
import pandas as pd
from functools import partial

def bisect(f, lb, ub, n = 0, aim = np.nan):
    """
    bisection like scipy.optimize.bisect but supports discontinous functions snapping towards our "aim"
    
    Example:
    --------
    >>> f = lambda x: 1 if x>0 else -1
    >>> assert bisect(f, 1, 3, aim = 1.3) == 1.3
    >>> assert bisect(f, -1, 3, aim = 1.3) == 1.3    

    >>> f = lambda x: 1 if x>2 else 0.5 if x>1 else -1 
    >>> assert round(bisect(f, -1, 3, 20, aim = 3),3) == 1.0 ## The crossing from positive to negative is at 1 so aim of 3 does not chain

    """
    flb = f(lb)
    fub = f(ub)
    for i in range(n or 10):
        if fub * flb >= 0 :
            break
        else:
            mid = (lb + ub) / 2
            fmid = f(mid)
            if flb * fmid < 0:
                ub = mid
                fub = fmid
            else:
                lb = mid
                flb = fmid
    mid = (lb + ub) / 2
    if np.isnan(aim):
        return mid                    
    elif f(aim) == f(mid):
        return aim
    elif aim < mid:
        return lb
    else:
        return ub


def _single_step_multibuffer(target, position_target, band, point_values, vol, correlations, previous, unit = 1.0, prev_m = 1.0, risk_band = 0.1, rounding_band = 0):    
    """
    Since actual positions are discrete while the target positions can be a fraction, 
    _single_step_multibuffer adjusts the target risk using a multiplier so that the realised (integer valued) buffered positions match the overall risk within [1-risk_band, 1+risk_band]
    
    :Parameters:
    ------------
    - target : array
        target for calculation of VaR
    - position_target
        target for calculation of position
    - band : array of floats
        bands for the target positions of individual markets (see buffer function)
    - point_values : array
        a multiplier per each position. This is useful if for example, a unit position in a market actually corresponds to much higher value (e.g. in Futures, point_values will be fx_rate * full_point_value)
    - vol: array
        price volatility
    - correlations: 
        actually, correlations between the markets
    - previous: array
        previous integer valued positions
    - unit: int
        The minimum amount of trading in an asset (e.g 1 future lots)
    - prev_m: float
        The previous adjustment multiplier
    - risk_band: float
        The band of risk we would like to keep overall
        
        
    :Example:
    ---------
    >>> correlations = np.array([[1,0.3], [0.3, 1]])
    >>> point_values = np.array([0.5, 0.5])
    >>> vol = np.array([1.2, 4.3])
    >>> mult = 3
    >>> target = np.array([0.3, 0.8]) ## current target positions
    >>> previous = np.array([0., 2.]) ## previous position
    >>> band = np.array([0.1, 0.1])
    >>> prev_m = 1.0
    >>> risk_band = 0.1
    
    
    >>> from pyg import *
    >>> a = np.array([target])
    >>> state = Dict(pos = list(previous))
    >>> u = list(unit)
    
    """
    pv = point_values * vol
    w = np.matmul(np.array([pv]).T, np.array([pv]))
    cw = correlations * w
    cw[np.isnan(cw)] = 0.0
    target[np.isnan(target)] = 0.0
    position_target[np.isnan(position_target)] = 0.0
    target_risk = matmul(cw, target, target, 0.5)
    b = np.array([band])
    state = dict(pos = list(previous))
        
    if target_risk == 0:
        def _mismatch(buffered_pos, target_mismatch = 1.0):
            return 0.
        def do_buffer_and_calculate_mismatch(m, target_mismatch = 1.0):
            return 0.
    else:        
        
        def _mismatch(buffered_pos, target_mismatch = 1.0):
            buffered_risk = matmul(cw, buffered_pos, buffered_pos, 0.5)
            return buffered_risk / target_risk - target_mismatch
        
        def do_buffer_and_calculate_mismatch(m, target_mismatch = 1.0):
            buffered_pos = buffer(a = np.array([position_target * m]), band = b, state = state, unit = unit, rounding_band = rounding_band)[0]
            return _mismatch(buffered_pos, target_mismatch)

    m = prev_m
    buffered_pos = buffer(a = np.array([position_target * m]), band = b, state = state, unit = unit, rounding_band = rounding_band)[0]
    mismatch = _mismatch(buffered_pos)

    if mismatch < - risk_band:
        function = partial(do_buffer_and_calculate_mismatch, target_mismatch = 1 - risk_band) ## we need to drive up risk but only to 1-risk_band
        m = bisect(function, lb = prev_m, ub = prev_m + risk_band, n = 5, aim = prev_m)        
        buffered_pos = buffer(a = np.array([position_target * m]), band = b, state = state, unit = unit)[0]
        # m = prev_m * (1 - risk_band) / mismatch
        mismatch = _mismatch(buffered_pos)
        if mismatch > risk_band:
            m = prev_m
            buffered_pos = buffer(a = np.array([position_target * m]), band = b, state = state, unit = unit, rounding_band = rounding_band)[0]
            

    elif mismatch > risk_band:
        function = partial(do_buffer_and_calculate_mismatch, target_mismatch = 1 + risk_band)
        m = bisect(function, lb = prev_m - risk_band, ub = prev_m, n = 5, aim = prev_m)                
        # m = prev_m * (1 + risk_band) / mismatch
        buffered_pos = buffer(a = np.array([position_target * m]), band = b, state = state, unit = unit, rounding_band = rounding_band)[0]
        mismatch = _mismatch(buffered_pos)
        if mismatch < - risk_band:
            m = prev_m
            buffered_pos = buffer(a = np.array([position_target * m]), band = b, state = state, unit = unit, rounding_band = rounding_band)[0]
    
    return m, buffered_pos, _mismatch(buffered_pos)


@loop(list, tuple, dict)
def _from(value, n, t = None):
    if isinstance(value, np.ndarray) and (t is None or len(value) == t):
        return value[n:]
    else:
        return value

def _subset(value, keys = None, ids = None):
    if keys is None and ids is None:
        return value
    elif is_df(value):
        return value.iloc[:, ids] if keys is None else value[keys]
    elif is_series(value):
        if is_ts(value):
            return value
        else:
            return value.iloc[ids] if keys is None else value[keys]
    elif is_arr(value):
        if ids is None:
            return value
        elif len(value.shape) == 1:
            return value[ids]
        elif len(value.shape) == 2:
            return value[:,ids]
        else:
            res = value[:,ids,:]
            res = res[:,:,ids]
            return res
    else:
         return value
    

def near_correlation_matrix(near_correlation, n):
    """
    creates an nxn array with correlation decaying as we leave the diagonal
    
    Example
    -------
    >>> near_correlation_matrix(0.5, 4)
    array([[1.   , 0.5  , 0.25 , 0.125],
           [0.5  , 1.   , 0.5  , 0.25 ],
           [0.25 , 0.5  , 1.   , 0.5  ],
           [0.125, 0.25 , 0.5  , 1.   ]])    

    """
    return np.array([[1 if i == j else near_correlation**abs(i-j) for i in range(n)] for j in range(n)])

def beta_correlation_matrix(beta, n = None):
    """
    creates an nxn array with correlation corresponding to exposure of each variable to a single factor
    
    Example
    -------
    >>> beta_correlation_matrix([0.9, 0.8, 0.5, -0.3])
    array([[ 1.  ,  0.72,  0.45, -0.27],
           [ 0.72,  1.  ,  0.4 , -0.24],
           [ 0.45,  0.4 ,  1.  , -0.15],
           [-0.27, -0.24, -0.15,  1.  ]])

    """
    if n is None:
        n = len(beta)
    if is_num(beta):
        beta = [beta] * n
    return np.array([[1 if i == j else beta[i]*beta[j] for i in range(n)] for j in range(n)])



@pd2np
def _multibuffer(target, band, unit, correlations, volatilities, point_values, position_target = None, data = None, mult = None, mismatch = None, risk_band = 0.1, rounding_band = 0):
    if position_target is None:
        position_target = target
    if data is None:
        previous = np.zeros(target.shape[1])
        m = 1
    else:
        previous = data[-1]
        m = mult[-1]
        n = len(data)
        t = len(target)
        if n == t:
            return Dict(mult = mult, data = data, mismatch = mismatch)
        target, position_target, band, unit, correlations, volatilities, point_values = _from([target, position_target, band, unit, correlations, volatilities, point_values], n = n, t = t)
    ds = np.zeros(target.shape)
    ms = np.ones(target.shape[0])
    mismatches = np.zeros(target.shape[0])
    t0 = dt()
    for i in range(target.shape[0]):
        m, previous, match = _single_step_multibuffer(target = target[i], 
                                                      position_target = position_target[i],
                                                      band = band if is_num(band) else band[i], 
                                                      rounding_band = rounding_band,
                                                      point_values = point_values[i], vol = volatilities[i], 
                                               correlations = correlations[i], previous = previous, unit = unit, prev_m = m, risk_band = risk_band)
        ds[i] = previous
        ms[i] = m
        mismatches[i] = match
        if i % 1000 == 0:
            t1 = dt()
            print(i, t1 - t0)
            t0 = t1
    if data is None:
        return Dict(mult = ms, data = ds, mismatch = mismatches)
    else:
        return Dict(mult = np.concatenate([mult, ms]), 
                    data = np.concatenate([data, ds]), 
                    mismatch = np.concatenate([mismatch, mismatches]))

_multibuffer.output = ['data', 'mult', 'mismatch']

def _subset_multibuffer(subset, target, band, unit, correlations, volatilities, point_values, position_target = None, data = None, mult = None, mismatch = None, risk_band = 0.1, rounding_band = 0):
    if is_nums(subset):
        ids = subset
        keys = target.columns[ids] if is_df(target) else None
    elif is_strs(subset):
        keys = subset
        ids = Dict(zip(target.columns, range(len(target.columns))))[tuple(keys)] if is_df(target) else None
    else:
        keys = ids = None
    variables = [target, position_target, band, unit, correlations, volatilities, point_values, data , mult, mismatch, risk_band, rounding_band]
    target, position_target, band, unit, correlations, volatilities, point_values, data , mult, mismatch, risk_band, rounding_band = [_subset(i, ids = ids, keys = keys) for i in variables]
    # target, band, unit, correlations, volatilities, point_values, data , mult, mismatch, risk_band, rounding_band = _subset([target, band, unit, correlations, volatilities, point_values, data, mult, mismatch, risk_band, rounding_band], ids = ids, keys = keys)
    return _multibuffer(target = target, position_target = position_target, band = band, unit = unit, correlations = correlations, volatilities = volatilities, point_values = point_values, data = data, mult = mult, mismatch = mismatch, risk_band = risk_band, rounding_band = rounding_band)


@loop(tuple)
def _to_target(value, target = None):
    """aligns value with target"""
    if isinstance(value, list) and min([is_ts(v) or is_num(v) for v in value]):
        value = df_concat(value)
    if is_df(target) and is_df(value):
        value = ffill(df_reindex(value, target))
    return ffill(value) if isinstance(value, (np.ndarray,pd.DataFrame))  else value
        

def multibuffer(target, band, unit, correlations, volatilities, point_values, position_target = None, data = None, mult = None, mismatch = None, 
                risk_band = 0.1, subset = None, subset_mult = None, subset_mismatch = None, correlations_index = None, rounding_band = 0):
    """
    performs a buffering of a target position but aiming to target a given level of risk
    
    TargetRisk is assumed to be given as 
    
    single_position_volatilities = target * point_values * volatilities
    portfolio_risk = single_position_volatilities^T x  correlations x single_position_volatilities

    first_attempt = buffer(target, band = band / (volatilities * point_values))
    
        
    :Parameters:
    ------------
    - target: df
        The target for the purpose of calculating RISK
        
    - position_target: df
        The target for the purpose of calculating POSITIONS. It seems odd that the two can be different. However...
        We may have run a process like multibuffer before on some of the data, arriving at a multiplier for a subset of the data
    
    - band: float/df
        The buffer on the position. See single buffer function for explanation
    
    - volatilities:
        timeseries of volatilities
    
    - point_values: 
        point values of each of the markets. The risk in each ticker is actually point_value * vol
        
    - unit: int
        rounding unit for the purpose of buffering
        
    - correlations:
        a 3d matrix of correlations or a 
    
    
    :Example:
    ---------
    >>> from pyg import *
    >>> target = get_data('data','long_only', item = 'lots') ## grab original target lots per each asset over time
    >>> band = get_data('data','long_only', item = 'band')
    >>> unit = 1.0
    >>> volatilities = get_data('data','long_only', item = 'fut_vol')
    >>> point_values = get_data('data','long_only', item = 'rpz')
    >>> correlations = get_data('data','long_only', item = 'covariance_zero')
    >>> buffered_lots = get_data('data','long_only', item = 'buffered_lots')
    >>> risk_band = 0.1
    >>> data = None ## output
    >>> mult = None ## output
    >>> subset = None
    
    >>> res = multibuffer(target, band, unit, correlations, volatilities, point_values, data = data, mult = mult, risk_band = risk_band, subset = subset)


    
    >>> res.mult.plot(title = 'adjustment over time to the risk')

    >>> print(dictable(ticker = list(res.data.columns), new = list(tover(res.data).values), b = list(tover(buffered_lots).values)).do(f12))
    >>> ticker = 'HGA Comdty'    
    >>> df_concat([target[ticker], res.data[ticker], buffered_lots[ticker]], ['fractional position', 'multi asset buffer', 'single asset buffer']).plot(title = ticker)
    
    
    In addition, you may want to ensure the risk in different sub-populations is maintained as well. 
    For this purpose, you can either run it with 
        - subset = [... list of all bonds]. This will subset the data to run on the bonds only
        - subset = dict(telecom = [list of telecom stocks], health = [list of health stocks], ...)
        This will return a multi-column mult dataframe with the multiplier per each sector and data = final positions for all stocks
                        
    :Example:
    ---------
    >>> res = multibuffer(target, band, unit, correlations, volatilities, point_values, data = data, mult = mult, risk_band = risk_band, subset = ['TYA Comdty', 'ESA Index'])
    
    
    """
    target = _to_target(target)
    volatilities, point_values, band, position_target = _to_target((volatilities, point_values, band, position_target), target)

    if is_num(correlations):
        correlations = near_correlation_matrix(correlations)
        print('near')
    if is_list(correlations):
        correlations = np.array(correlations)
        print('correlation list')
    if len(correlations.shape) == 1:
        print('beta correlation')
        correlations = beta_correlation_matrix(correlations)
    if is_df(target): 
        if len(correlations.shape) == 3 and correlations_index is not None and len(correlations) == len(correlations_index):
            correlations = reindex_3d(correlations, index = target.index, original_index = correlations_index)
        if len(correlations.shape) == 2:
            print('constant correlation, now in 3d')
            correlations = np.array([correlations] * len(target))
    if isinstance(subset, dict):
        res = {}
        for key in subset:
            value = subset[key]
            print('running', key, 'for', value)
            assert position_target is None
            res[key] = _subset_multibuffer(subset = value, 
                                           target = target, 
                                           band = band, 
                                           unit = unit, 
                                           correlations = correlations, 
                                           volatilities = volatilities, 
                                           point_values = point_values, 
                                           position_target = position_target,
                                           data = data, 
                                           mult = subset_mult[key] if is_df(subset_mult) else mult, 
                                           mismatch = subset_mismatch[key] if is_df(subset_mismatch) else subset_mismatch, 
                                           risk_band = risk_band,
                                           rounding_band = rounding_band)
        rtn = Dict(subset_mult = df_concat([r.mult for r in res.values()], list(res.keys())),
                   subset_mismatch = df_concat([r.mismatch for r in res.values()], list(res.keys())),
                   data = df_concat([r.data for r in res.values()]),
                   mult = df_concat(sum([[res[s].mult] * len(subset[s]) for s in subset], []), sum(subset.values(), [])),
                   mismatch = df_concat(sum([[res[s].mismatch] * len(subset[s]) for s in subset], []), sum(subset.values(), []))
                   )
        rtn['data'] = rtn['data'][target.columns]
        rtn['mult'] = rtn['mult'][target.columns]
        rtn['mismatch'] = rtn['mismatch'][target.columns]
        return rtn
    else:
        res = _subset_multibuffer(subset = subset, target = target, position_target = position_target, band = band, unit = unit, correlations = correlations, volatilities = volatilities, point_values = point_values, 
                              data = data, mult = mult, mismatch = mismatch, risk_band = risk_band, rounding_band = rounding_band)
        res['subset_mult'] = None
        res['subset_mismatch'] = None
    return res

multibuffer.output = _multibuffer.output + ['subset_mult', 'subset_mismatch']




import pandas as pd
import numpy as np
from scipy.optimize import minimize
TOLERANCE = 1e-10
from pyg_base import pd2np, is_df, skip_if_data_pd_or_np, try_nan, as_list, reducer, is_num, is_nums, is_pd, is_tss, is_list, df_concat, df_reindex
from pyg_timeseries._rolling import v2na, na2v, ffill
from pyg_timeseries._decorators import mask_nans, apply_along_first_axis


@skip_if_data_pd_or_np
@pd2np
@apply_along_first_axis(base_shape = 2)
@mask_nans
def _eigenvalues(matrix, sort = -1):
    try:
        evals, evec = np.linalg.eig(matrix)
    except Exception:
        return np.full(matrix.shape[0], np.nan)
    if sort!=0 and len(evals):
        _, ns = zip(*sorted(zip(evals, range(len(evals))))[::sort])
        ns = np.array(ns)
        evals = evals[ns]
    return evals    

    
def eigenvalues(matrix, index = None, data = None):
    """
    A thin wrapper for numpy np.linalg.eig
    By default, eigen values (and vectors) are sorted in decreasing order
    
    :Parameters:
    ------------
    matrix:
        A timeseries (t x n x n numpy array) or a single n x n matrix 
    index: 
        The result can be put into a timeseries if index is provided
    sort: int
        sorting mechanism for size of eigenvalues

    """
    return _eigenvalues(matrix, index = index, data = data)


@skip_if_data_pd_or_np
@pd2np
@apply_along_first_axis(base_shape = 2)
@mask_nans
def _eigenvectors(matrix, n = None, sort = -1):
    try:
        evals, evec = np.linalg.eig(matrix)
    except Exception:
        return np.full(matrix.shape if n is None else matrix.shape[0], np.nan)
    if sort!=0 and len(evals):
        _, ns = zip(*sorted(zip(evals, range(len(evals))))[::sort])
        ns = np.array(ns)
        evec = evec[:,ns]        
    if n is None:
        return evec
    else:
        if n < len(evec):
            return evec[:, n]
        else:
            return np.full(len(evec), np.nan)
    
def eigenvectors(matrix, n = None, sort = -1, columns = None, index = None, data = None):
    """
    A thin wrapper for numpy np.linalg.eig
    By default, eigen values (and vectors) are sorted in decreasing order
    
    :Parameters:
    ------------
    matrix:
        A timeseries (t x n x n numpy array) or a single n x n matrix 
    n: int or None
        if provided, will return the nth eigen vector (by sorted size)
    
    index: list/pd.Index of dates
        The result can be put into a timeseries if index is provided

    columns: list of components for matrix
        If n is provided, the result can be put into a timeseries dataframe if index is provided. Columns represent the projection of the eigen-vector

    sort: int
        sorting mechanism for size of eigenvalues

    """
    return _eigenvectors(matrix, n = n, columns = columns, index = index, data = data)


@skip_if_data_pd_or_np
@pd2np
@apply_along_first_axis(base_shape = 2)
@mask_nans
@try_nan
def _det(matrix):
    # if len(matrix.shape) == 1 and int(np.sqrt(matrix.shape[0]))**2 == matrix.shape[0]:
    #     n = int(np.sqrt(matrix.shape[0]))
    #     matrix = matrix.reshape((n,n))
    return np.linalg.det(matrix)    

def det(matrix, data = None):
    """
    A thin wrapper around np.linalg.det.
    
    - handles a timeseries of matrices
    - handles nan's a little better

    :Example:
    ---------
    >>> matrix = np.array([[1,2],[3,4]])
    >>> assert round(det(matrix),10) == -2.

    >>> matrices = np.array([matrix] * 5)   ### 5 x 2 x 2 is viewed as a timeseries of matrices
    >>> assert len(det(matrices)) == 5      ### get 5 x 1 back

    >>> matrix = np.array([[1,np.nan],[np.nan,np.nan]])     ## we drop rows and columns (2nd ones here) that are all nan
    >>> assert det(matrix) == 1
    """
    return _det(matrix, data = data)



from pyg_timeseries import diff, ratio
from pyg_timeseries import rolling_max, rolling_mean, rolling_median, rolling_quantile, rolling_rank, rolling_rms, rolling_skew, rolling_std, rolling_sum
from pyg_base import eq, drange
import pandas as pd
import numpy as np

def test_diff_with_time():
    a = 1. * np.arange(10)**2
    time = np.array([0,0,1,1,1,2,2,2,2,3])
    dff = np.array([np.nan, np.nan, 4-1, 9-1, 16-1, 25-16, 36-16, 49-16, 64-16, 81-64])
    assert eq(dff, diff(a, n=1, time = time))    
    rto = np.array([np.nan, np.nan, 4/1, 9/1, 16/1, 25/16, 36/16, 49/16, 64/16, 81/64])
    assert eq(rto, ratio(a, n=1, time = time))    
    dff2 = np.array([np.nan, np.nan, np.nan, np.nan, np.nan, 25-1, 36-1, 49-1, 64-1, 81-16])
    assert eq(dff2, diff(a, n=2, time = time))    
    rto2 = np.array([np.nan, np.nan, np.nan, np.nan, np.nan, 25/1, 36/1, 49/1, 64/1, 81/16])
    assert eq(rto2, ratio(a, n=2, time = time))    


def test_diff():
    a = 1. * np.arange(10)**2
    state = 0.
    assert diff(a, state = state)[0] == 0

    b = pd.Series(a, drange(9))
    assert eq(diff(b), b.diff())
    assert diff(b, state = state).iloc[0] == 0

    d = pd.DataFrame(dict(a = a), drange(9))
    assert eq(diff(d), d.diff())


def test_rolling_empty():
    for a in (np.array([]), pd.Series([],[], dtype = float), pd.DataFrame([],[])):
        for f in [rolling_max, rolling_mean, rolling_median, rolling_quantile, rolling_rank, rolling_rms, rolling_skew, rolling_std, rolling_sum]:
            res = f(a, 20)
            assert eq(res, a)



from pyg_base import df_reindex, df_concat, is_nums, Dict, mul_, loop, dictable, try_back, is_df, pd2np
from pyg_timeseries._rolling import na2v, ffill, v2na, fnna, shift
from pyg_timeseries._ewm import ewma, ewmcorr, ewmstd
from pyg_timeseries._pandas import fnna_like
import numpy as np
import pandas as pd

    
@pd2np
def _ewmcombine(a, w, n = 128, vol_days = None, corr = None):
    """
    We assume all the joining together etc. is done
    
    Parameters
    ----------
    a : np.array
        2-dimensional array of signals
    w : np.ndarray
        2-dimensional array of weights

    n : int
        days for which we smooth the correlation estimate
        
    Returns
    -------
    

    :Math:
    ------
    
    We have a collection of signals that we assume are RMS-1, i.e. N(0,1)-ish.
    Our relative weights w on each of the signals changes over time, perhaps because 
    - we allocate differently or 
    - because some signals only live at certain times of the year.


    We would like to create a variable X:
    X = \sum_i w_i * a_i

    But we also want to scale it so that X is RMS-1
    Since X is not necessarily RMS-1, we would want to divide it by its own vol estimate, which we can do with ewmrms(X, n)
    
    However, this does not take into account the fact that weights can change over time so the vol estimate yesterday may be irrelevant as w have changed.


    Single factor model
    -------------------        
    So let us assume a single factor correlation rho(t)

    E(X) = 0
    E(X^2) = X2 = \sum w_i^2 + \sum_{i<>j} w_i w_j rho(t)
        
    W1 = \sum w_i
    W2 = \sum w_i^2
    WIJ =  \sum_{i<>j} w_i w_j = \sum(i,j) w_i w_j - W2 = W1**2 - W2
    
    We reverse the equation and express rho as a function of E(X^2) 
    rho(t) = (X2 - W2)/WIJ
    
    This is a point estimate for rho. Once we have rho, we can estimate the unknown variance of X at each point of time.

    Full factor model
    -----------------
    Here we use ewmcorr to estimate a running correlation between all the signals. If we don't have a value (e.g. this is a recent timeseries so no data as yet)
    then we default to rho(t) from the single factor model above.
    
    Once we have the full correlation matrix estimation, we can estimate variance as...
    E(X^2) = w^T * C * w

    """
    fs = fnna(a)
    a_ = na2v(ffill(a))
    w_ = na2v(ffill(w))
    for i, j in enumerate(fs):
        w_[:j, i] = 0
    if vol_days:
        vols = v2na(ffill(ewmstd(a, vol_days)))
        a_ = a_ / vols
    else:
        vols = 1
    x = np.sum(a_ * w_, axis = 1)
    w1 = np.sum(w_, axis = 1)
    x2 = x**2
    w2 = np.sum(w_ ** 2, axis = 1)
    wij = w1 ** 2 - w2
    wij[wij == 0] = np.nan
    rho = (x2 - w2)/wij
    erho = ewma(rho, n) # this gives us our estimate for rho at time t
    if corr is True: 
        corr = n
    elif corr is False:
        corr = None
    full = corr is not None
    if full:
        c = ewmcorr(a, corr) if isinstance(corr, int) else corr
        c_ = np.zeros(c.shape)
        variance = np.empty(c.shape[0])
        cip = np.zeros(c.shape[1:])
        for i in range(c.shape[0]):
            ci = c[i]
            wi = w_[i]
            ci[np.isnan(ci)] = cip[np.isnan(ci)]
            ci[np.isnan(ci)] = erho[i]
            variance[i] = np.matmul(wi, np.matmul(ci,wi))
            c_[i] = ci
    else:
        variance = w2 + wij * erho # and hence an estimate for variance
    variance[variance<=0] = np.nan
    variance = shift(variance)
    vol = ffill(np.sqrt(variance))
    mult = 1/vol
    data = x/vol
    normalized_mult = mult * w1
    return dict(data = x, vol = vol, rho = erho, mult = mult, normalized_mult = normalized_mult, cor = c_ if full else None, vols = vols if vol_days else 1)


@loop(list)
def _col(value, i):
    if isinstance(value, np.ndarray) and len(value.shape) > 1:
        return value[:, i] if value.shape[1]>1 else value[:, 0]
    elif isinstance(value, pd.DataFrame):
        return value.iloc[:, i] if value.shape[1]>1 else value.iloc[:,0]
    else:
        return value        
    

def ewmcombine(tss, wgts, n = 128, vol_days = None, corr = None, join = 'oj', method = None):
    """
    We expect tss and wgts to be provided as lists at the moment

    Parameters
    ----------
    tss : list of timeseries
        the data we want to combine into a single signal
    wgts : list/dataframe of weights, either floats or timeseries
        
    n : int, optional
        Length of data for correlation calculation. 
    corr : a 3-d numpy matrix. optional
        The output from ewmcorr. If provided, a full calculation of correlation is used
    vol_days:
        The input assumes that tss are broadly N(0,1) if this is false, the correlation calculation will be off. 
        You can set vol_days and we will normalize each tss on the fly. If so, the normalized data will be available in vols output
    join : str, optional
        how we join the tss to a single frame. The default is 'oj' for outer join.
    method : TYPE, optional
        how we fill nans when we join the tss to a single frame. The default is None .

    Returns
    -------
    res : TYPE
        DESCRIPTION.

    """
    if is_df(tss):
        wgts = fnna_like(tss, wgts).ffill()
        return _ewmcombine(tss, wgts, n = n, vol_days = vol_days, corr = corr)      
    if len(tss) == 1:
        return Dict(data = tss[0], vol = None, rho = None, mult = 1/wgts[0], normalized_mult = 1, cor = None, vols = 1)            
    cols = [tuple(ts.columns) for ts in tss if is_df(ts) and ts.shape[1] > 1]
    if len(cols):
        assert len(set(cols)) == 1
        cols = list(cols[0])
        res = Dict(dictable([ewmcombine(_col(tss, i), _col(wgts, i), n = n, corr = corr, join = join, method = method) for i in range(len(cols))]))
        res = res.do(try_back(lambda v: df_concat(v, cols)))
        return res        
    if is_nums(wgts):
        tss, wgts = zip(*[(ts if w>0 else -ts, abs(w)) for ts, w in zip(tss, wgts) if w!=0]) ## we drop any zero weights and invert any negative weights
        a = df_concat(list(tss))
        w = fnna_like(a, wgts, 0.).values.astype(float)
    else:
        tss, wgts = zip(*[(mul_(ts,np.sign(df_reindex(w, ts, 'ffill'))), abs(w)) for ts, w in zip(tss, wgts)]) ## if we is a timeseries, we multiply the two after reindexing w to ts
        a = df_concat(list(tss))
        w = df_concat([df_reindex(wgt, a.index, method = 'ffill') for wgt in wgts]).values.astype(float)
    if len(tss) == 1:
        return Dict(data = tss[0], vol = None, rho = None, mult = 1/wgts[0], normalized_mult = 1, cor = None, vols = 1)            
    idx = a.index
    res = _ewmcombine(a.values, w, n = n, vol_days = vol_days, corr = corr)
    res = Dict({k : (pd.Series(v, idx) if len(v.shape) == 1 else [pd.Series(v[:,i], idx) for i in range(v.shape[1])]) if isinstance(v, np.ndarray) and v.shape[0] == len(idx) else v for k, v in res.items()})
    return res

ewmcombine.output = ['data', 'vol', 'rho', 'mult', 'normalized_mult',  'cor', 'vols']

def ewmcombined(tss, wgts, n = 128, vol_days = None, corr = False, join = 'oj', method = None):
    """
    We expect tss and wgts to be provided as lists at the moment

    Parameters
    ----------
    tss : list of timeseries
        the data we want to combine into a single signal
    wgts : list/dataframe of weights, either floats or timeseries
        
    n : int, optional
        Length of data for correlation calculation. 
    corr : a 3-d numpy matrix. optional
        The output from ewmcorr. If provided, a full calculation of correlation is used
    vol_days:
        The input assumes that tss are broadly N(0,1) if this is false, the correlation calculation will be off. 
        You can set vol_days and we will normalize each tss on the fly. If so, the normalized data will be available in vols output
    join : str, optional
        how we join the tss to a single frame. The default is 'oj' for outer join.
    method : TYPE, optional
        how we fill nans when we join the tss to a single frame. The default is None .

    Returns
    -------
    res : TYPE
        DESCRIPTION.

    """
    return ewmcombine(tss, wgts, n = n, corr = corr, join = join, method = method)['data']

    